#
# Copyright 2019 GridGain Systems, Inc. and Contributors.
#
# Licensed under the GridGain Community Edition License (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gridgain.com/products/software/community-edition/gridgain-community-edition-license
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
TYPE_BYTE = 1
TYPE_SHORT = 2
TYPE_INT = 3
TYPE_LONG = 4
TYPE_FLOAT = 5
TYPE_DOUBLE = 6
TYPE_CHAR = 7
TYPE_BOOLEAN = 8
TYPE_STRING = 9
TYPE_UUID = 10
TYPE_DATE = 11
TYPE_BYTE_ARR = 12
TYPE_SHORT_ARR = 13
TYPE_INT_ARR = 14
TYPE_LONG_ARR = 15
TYPE_FLOAT_ARR = 16
TYPE_DOUBLE_ARR = 17
TYPE_CHAR_ARR = 18
TYPE_BOOLEAN_ARR = 19
TYPE_STRING_ARR = 20
TYPE_UUID_ARR = 21
TYPE_DATE_ARR = 22
TYPE_OBJ_ARR = 23
TYPE_COL = 24
TYPE_MAP = 25
TYPE_BINARY_OBJ = 27
TYPE_ENUM = 28
TYPE_ENUM_ARR = 29
TYPE_DECIMAL = 30
TYPE_DECIMAL_ARR = 31
TYPE_CLASS = 32
TYPE_TIMESTAMP = 33
TYPE_TIMESTAMP_ARR = 34
TYPE_PROXY = 35
TYPE_TIME = 36
TYPE_TIME_ARR = 37
TYPE_BINARY_ENUM = 38
