#
# Copyright 2019 GridGain Systems, Inc. and Contributors.
#
# Licensed under the GridGain Community Edition License (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gridgain.com/products/software/community-edition/gridgain-community-edition-license
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
TC_BYTE = b'\x01'
TC_SHORT = b'\x02'
TC_INT = b'\x03'
TC_LONG = b'\x04'
TC_FLOAT = b'\x05'
TC_DOUBLE = b'\x06'
TC_CHAR = b'\x07'
TC_BOOL = b'\x08'
TC_STRING = b'\x09'
TC_UUID = b'\x0a'
TC_DATE = b'\x0b'
TC_BYTE_ARRAY = b'\x0c'
TC_SHORT_ARRAY = b'\x0d'
TC_INT_ARRAY = b'\x0e'
TC_LONG_ARRAY = b'\x0f'
TC_FLOAT_ARRAY = b'\x10'
TC_DOUBLE_ARRAY = b'\x11'
TC_CHAR_ARRAY = b'\x12'
TC_BOOL_ARRAY = b'\x13'
TC_STRING_ARRAY = b'\x14'
TC_UUID_ARRAY = b'\x15'
TC_DATE_ARRAY = b'\x16'
TC_OBJECT_ARRAY = b'\x17'
TC_COLLECTION = b'\x18'
TC_MAP = b'\x19'

TC_ARRAY_WRAPPED_OBJECTS = b'\x1b'

TC_ENUM = b'\x1c'
TC_ENUM_ARRAY = b'\x1d'

TC_DECIMAL = b'\x1e'
TC_DECIMAL_ARRAY = b'\x1f'
TC_TIMESTAMP = b'\x21'
TC_TIMESTAMP_ARRAY = b'\x22'

TC_TIME = b'\x24'
TC_TIME_ARRAY = b'\x25'
TC_BINARY_ENUM = b'\x26'

TC_NULL = b'\x65'
TC_COMPLEX_OBJECT = b'\x67'
